{==========================================================================}
{ Program            : MOSSIM                                              }
{ Author             : Robert Lacoste, Chaville, France                    }
{ Version            : 1.3                                                 }
{ Release date       : 27/07/1996                                          }
{==========================================================================}

{--------------------------------------------------------------------------}
{                              UNIT SIMUL                                  }
{                                                                          }
{ Description : Calculate the outputs signals with the given inputs        }
{--------------------------------------------------------------------------}

{++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
Principle :

For each time step, the following 4 phases are processed until a stability
is reached :

ph 1/ Groups of nodes are elaborated, based of connections between nodes by
      a transistor with an active gate level
ph 2/ "Uncertain links" are stored when two groups are connected by a
      transistor with a gate at an uncertain state
ph 3/ Groups states are calculated given the states of each nodes
ph 4/ Effects of uncertain links are calculated

Each node, or group, can have one of 6 logical states :
- Forced 0 (vss or input defined at 0
- Forced 1 (vcc or input defined at 1
- Pullup
- High impedance, state 1
- High impedance, state 0
- High impedance, undefined state

The rules for phase 3 (group states calculation) are the following
(by decreasing priority) :

Rule 1 : If two nodes of a same group are forced to different states, the
         simulation is stopped (conception error)
Rule 2 : If one node of a same group is forced 1 or forced 0, the group state
         is the same
Rule 3 : If one node of a group have a pullup, the group state is pullup
Rule 4 : If all nodes of a group are high impedance and if the number of
         nodes in state 1 is more than state 0 or undefined, group state
         is high impedance, state 1
Rule 5 : If all nodes of a group are high impedance and if the number of
         nodes in state 0 is more than state 1 or undefined, group state
         is high impedance, state 0
If no rule apply, the group state is undefined

The rules for phase 4 (uncertain links, or UL, resolution) are the following :

Rule 1 : If two groups connected by an UL are forced to different states,
         the simulation is stopped (conception error)
Rule 2 : If a group A, state pullup or high impedance 1, is connected by
         an UL to a group forced 0, group A state is uncertain
Rule 3 : If a group A, state high impedance 0, is connected by an UL to
         a group forced 1, group A state is uncertain
Rule 4 : If two groups A and B, high impedance states 0 and 1 respectively,
         are connected by an UL, the two groups states are uncertain

+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

UNIT SIMUL;

INTERFACE                       {external definitions}
  USES GRAPH,
       DATA,
       GDS,
       PROCGEN;

  {go the simulation}
  PROCEDURE SIMULATION(nbsteps:integer);

IMPLEMENTATION                  {local definitions}

  const
    NBMGROUPES = NBMNOEUDS;     {max number of groups}

  type
    pnnd= ^nnd;
	nnd= record                 {linked list of nodes}
		   q:zonenoeud;
		   nndsuiv:pnnd;
	     end;

    {electrical level of a group : forced 0, 1, pullup, normal O, 1 undef}
    niveauxgroupes =(e0,e1,pu,n0,n1,x);

    groupe= record              {group structure}
		      niveau: niveauxgroupes;
		      debgr: pnnd;
    	    end;

    plienx= ^lienx;			{linked list of uncertain links between nodes}
    lienx  = record
	       	   nd,nf: zonenoeud;       {first and second nodes numbers}
		       lxsuiv:plienx;          {next record}
	         end;

    zonegroupe= 0..NBMGROUPES;   {group number}


  var
      s:string;
      INDICE,x1,y1,nombreiterations,nombremodifs:integer;
      erreurtrouvee:boolean;

      tabletats:array[1..NBMNOEUDS] of tristate;	{state of each node}
      groupenoeud:array[1..NBMNOEUDS] of zonegroupe;{group of each node}
      groupes:array[1..NBMGROUPES] of groupe;		{groups array}
      lx,debutliensx,tx,ux: plienx;			        {list of uncertain links}

      u,t: pnnd;
      simheap:^INTEGER;

  {---------------- concatenation of the ndoes in groups ------------------}

  PROCEDURE FAITGROUPES;
  VAR
    TR,A1,A2:INTEGER;
    P,P1:PNND;
    LX:PLIENX;
    CONNECTES:TRISTATE;
  BEGIN

    {check each transistor connecting two nodes}
    for tr:=1 to NBMTRANSIS do if (transistors[tr]<>nil) then
      begin

        {get the numbers of the two nodes linked by this transistor}
        a1:=transistors[tr]^.t1; a2:=transistors[tr]^.t2;

        {given grid electrical level, check if transistor is connected}
        case noeuds[transistors[tr]^.grille]^.etat of
          un :     if transistors[tr]^.channel=nchannel then
                     connectes:=un else connectes:=zero;
          zero :   if transistors[tr]^.channel=pchannel then
                     connectes:=un else connectes:=zero;
          nondef : connectes:=nondef;
        end;

        case connectes of

          {if transistor is connected, the two nodes are in the same group}
          un   :begin

                  {if two already existent groups, merge them}
                  if ((groupenoeud[a1]<>0) and (groupenoeud[a2]<>0))
                  then begin
                         p:=groupes[groupenoeud[a1]].debgr;
                         while (p^.nndsuiv<>nil) do p:=p^.nndsuiv;
                         p^.nndsuiv:=groupes[groupenoeud[a2]].debgr;
                         groupes[groupenoeud[a2]].debgr:=nil;
                         groupenoeud[a2]:=groupenoeud[a1];
                       end

                  {if one is part of an existing group, add the second}
                  else if (groupenoeud[a1]<>0) then
                       begin
                         new(p);
                         p^.q:=a2;
                         p^.nndsuiv:=groupes[groupenoeud[a1]].debgr;
                         groupes[groupenoeud[a1]].debgr:=p;
                         groupenoeud[a2]:=groupenoeud[a1];
                       end
                  else if (groupenoeud[a2]<>0) then
                       begin
                         new(p);
                         p^.q:=a1;
                         p^.nndsuiv:=groupes[groupenoeud[a2]].debgr;
                         groupes[groupenoeud[a2]].debgr:=p;
                         groupenoeud[a1]:=groupenoeud[a2];
                       end

                  {else create a new group with the two nodes}
                  else
                    begin
                      new(p);
                      p^.q:=a1;
                      new(p1);
                      p1^.q:=a2;
                      groupes[a1].debgr:=p;
                      p^.nndsuiv:=p1;
                      p1^.nndsuiv:=nil;
                      groupenoeud[a1]:=a1;
                      groupenoeud[a2]:=a1;
                    end
                 end;

          {if gate is level undefined, create an uncertain link between}
          {the two groups}
          nondef:begin
                   {create groups if not already existents}
                   if (groupenoeud[a1]=0) then begin
                                             new(p); p^.q:=a1;
                                             p^.nndsuiv:=nil;
                                             groupes[a1].debgr:=p;
                                             groupenoeud[a1]:=a1;
                                           end;
                   if (groupenoeud[a2]=0) then begin
                                             new(p); p^.q:=a2;
                                             p^.nndsuiv:=nil;
                                             groupes[a2].debgr:=p;
                                             groupenoeud[a2]:=a2;
                                           end;

                   {add an uncertain link to the linked list}
                   new(lx);
                   lx^.nd:=a1; lx^.nf:=a2;
                   lx^.lxsuiv:=debutliensx;
                   debutliensx:=lx;
                 end;

          {if transistor is not connected, no connection}
          zero  : ;
        end;
      end;
    end;

   {------- determination of the electrical level of a given group -------}

   procedure calculetatgroupe(J:ZONEGROUPE);
    var nvcc,nvss,ne1,ne0,npull,nn1,nn0,nnx:integer;
	    p:pnnd; a:zonenoeud;
    begin

      {initialise the nodes states counters}
      nvcc:=0; nvss:=0; ne1:=0; ne0:=0; npull:=0; nn1:=0; nn0:=0; nnx:=0;

      {scan the linked list of nodes for this group}
      p:=groupes[j].debgr;
      while (p<>nil) do
	    begin

          {count the number of nodes for each electrical level}
          a:=p^.q;
	      case noeuds[a]^.classe of
	        vcc: nvcc:=nvcc+1;
	        vss: nvss:=nvss+1;
	        normal:
              begin
		        if (noeuds[a]^.xpull<>0) then npull:=npull+1;
		        case noeuds[a]^.etat of
			      zero: nn0:=nn0+1;
			      un  : nn1:=nn1+1;
			      nondef: nnx:=nnx+1;
		        end;
		      end;
	        entree:
              begin
		        case valentrees[noeuds[a]^.numentree,indice] of
			      0 : ne0:=ne0+1;
			      1 : ne1:=ne1+1;
		        end;
		      end;
	      end;
	      p:=p^.nndsuiv;
	    end;

      {check if incompatible electrical levels are present in this group}
      if (((nvcc<>0) or (ne1<>0)) and ((nvss<>0) or (ne0<>0))) then
		begin
          {if so, stop the calculation}
		  erreurtrouvee:=true;
	      groupes[j].niveau:=x;
		end

      {if at least one node is forced 1, group level is 1}
      else if ((nvcc<>0) or (ne1<>0)) then groupes[j].niveau:=e1

      {if at least one node is forced 0, group level is 0}
      else if ((nvss<>0) or (ne0<>0)) then groupes[j].niveau:=e0

      {if at least one node have a pullup, group level is pullup}
      else if (npull<>0) then groupes[j].niveau:=pu

      {if more nodes are level 1 than level 0 or undefined, group level is 1}
      else if (nn1-nn0-nnx>0) then groupes[j].niveau:=n1

      {if more nodes are level 0 than level 1 or undefined, group level is O}
      else if (nn0-nn1-nnx>0) then groupes[j].niveau:=n0

      {else group level is undefined}
      else groupes[j].niveau:=x;
    end;

 {-------------- computing effects of uncertain links ---------------------}

 procedure resoudx(D,F:ZONEGROUPE);
    var w1,w2: niveauxgroupes;
    begin

      {get electrical levels of the two groups}
      w1:=groupes[d].niveau; w2:=groupes[f].niveau;

      {if there is an uncertain incompatibility, stop the simulation}
      if (((w1=e0) and (w2=e1)) or ((w1=e1) and (w2=e0))) then
	      begin
		    erreurtrouvee:=true;
		    groupes[d].niveau:=x; groupes[f].niveau:=x;
	      end

      {cases where one group level become uncertain}
      else if ((w1=e0) and ((w2=pu) or (w2=n1))) then groupes[f].niveau:=x
      else if ((w1=e1) and (w2=n0)) then groupes[f].niveau:=x
      else if ((w2=e0) and ((w1=pu) or (w1=n1))) then groupes[d].niveau:=x
      else if ((w2=e1) and (w1=n0)) then groupes[d].niveau:=x
      else if ((w1=pu) and (w2=n0)) then groupes[f].niveau:=x
      else if ((w2=pu) and (w1=n0)) then groupes[d].niveau:=x
      else if (((w1=n0) and (w2=n1)) or ((w1=n1) and (w2=n0))) then
	    begin
		  groupes[d].niveau:=x; groupes[f].niveau:=x;
        end;
    end;

  {-------- calculate state of each node given group's states ------------}

  PROCEDURE METETATS;
    BEGIN

      {scan each node}
      for j:=1 to NBMNOEUDS do if (noeuds[j]<>nil) then

        {if this node is part of a group, node level = group level}
	    if (groupenoeud[j]<>0) then
		  begin
		    if (groupes[groupenoeud[j]].niveau in [e1,pu,n1])
		    then tabletats[j]:= un
		    else if (groupes[groupenoeud[j]].niveau=x)
			  then tabletats[j]:=nondef
			  else tabletats[j]:=zero;
		 end

        {if not, forced nodes have a defined value, and non connected nodes}
        {keep their previous value}
		else
		 begin
		   if (noeuds[j]^.classe=entree) then
			begin
			  if (valentrees[noeuds[j]^.numentree,indice] = 0)
			    then tabletats[j]:=zero
			    else tabletats[j]:=un;
			end
		   else if (noeuds[j]^.classe=vcc) then tabletats[j]:=un
		   else if (noeuds[j]^.classe=vss) then tabletats[j]:=zero
		   else if (noeuds[j]^.xpull<>0) then tabletats[j]:=un
		   else tabletats[j]:=noeuds[j]^.etat;
		 end;

      {copy nodes state and count nodes levels changes}
      for j:=1 to NBMNOEUDS do if (noeuds[j]<>nil) then
	   if (noeuds[j]^.etat<>tabletats[j]) then
	     begin
	       nombremodifs:=nombremodifs+1;
	       noeuds[j]^.etat:=tabletats[j]
	     end;
  END;

  {------------------ Initialise all nodes status to undefined ------------}

  PROCEDURE INITPHASEZERO;
  var i:integer;
  begin
    for i:=1 to NBMNOEUDS do if (noeuds[i]<>nil) then noeuds[i]^.etat:=nondef;
  end;

  {------------------ Calculate nodes states for one time step ------------}

  function calculphase:boolean ;
  begin

    nombreiterations:=0; erreurtrouvee:=false;

    {loop until stability}
    repeat
      nombremodifs:=0;

      {initialise states and groups numbers}
      for j:=1 to NBMNOEUDS do
      begin
        tabletats[j]:=nondef;groupenoeud[j]:=0;
      end;
      for j:=1 to NBMGROUPES do groupes[j].debgr:=nil;
      debutliensx:=nil;

      {find groups and uncertain links}
      faitgroupes;

      {calculate theorical state of each group}
      for j:=1 to NBMGROUPES do if groupes[j].debgr<>nil then
			calculetatgroupe(j);

      {recalculate states of nodes connected by an uncertain link}
      lx:=debutliensx;
      while (lx<>nil) do
        begin
		  resoudx(lx^.nd,lx^.nf);
          lx:=lx^.lxsuiv;
        end;

      {copy nodes states}
      METETATS;
      nombreiterations:=nombreiterations+1;

    until (erreurtrouvee or ((nombremodifs=0) or (nombreiterations>3*nbnoeud)));

    {check if a stability was found}
    if (nombreiterations>3*nbnoeud) then
      begin
		erreur(60);
		for j:=1 to NBMSORTIES do
					     valsorties[j,indice]:=nondef;
	 end

    {check if a conceptionerror was found}
    else if erreurtrouvee then
      erreur(61)

    {copy states of nodes connected to an output to this output}
	else for j:=1 to NBMNOEUDS do if (noeuds[j]<>nil) then
	  if (noeuds[j]^.numsortie<>0) then
	     valsorties[noeuds[j]^.numsortie,indice]:=noeuds[j]^.etat;

    {return true if ok}
    calculphase:=not(erreurtrouvee or (nombreiterations>3*nbnoeud));
  end;

  {---------------------- calculate states for all time steps ------------}

  procedure simulation{(nbsteps:integer)};
  var  i:integer; f:boolean;
  begin
    {mark memory head}
    MARK(SIMHEAP);

    {check if a recalculation is needed}
    if (nbnoeud <>0) then
    begin

      {initialise nodes states}
      initphasezero;

      {calculate states for each time step}
      i:=1;
      repeat
        {display the current time step number}
        CHAINEINT(indice,s);
        s:=concat('phase: ',s);
        info(s,WHITE);

	    INDICE:=I;
        f:=calculphase;
	    i:=i+1;
      until ((i>nbsteps) or not(f));
    end;

    effaceinfo;

    {release memory}
    RELEASE(SIMHEAP);
  end;

{---------------------------- Unit initialisation -------------------------}

BEGIN
END.


